package capsule;

import node.*;
import node.services.*;
import clientserver.*;

import java.io.*;
import java.util.*;

public class PingListenerAgent extends MyAgent {
    // Assumes ping packets as specified in icmp rfc 792
    protected final static int serieLen=10; // The number of pings in a serie.
    protected final static int pingDistance=10; // The time between two pings in 100 ms
    protected final static int pingTimeOut=10; // How long I wait after the last ping in 100 ms
    protected final static int syncOffset=5; // To sync multiple agents. Only every th ping can start a new serie.

    protected int series=5; // The number of series to be measured.


    protected PingSeriesCounter requestCntr;
    protected PingSeriesCounter replyCntr;
    protected int replyAnomaly;

    protected int identifier; // Only consider pings with the same identifier.

    protected boolean restart=true; // to signal that a new serie can be started.
    protected IPPacket curPacket;

    public void handlePacket(IPPacket p) {
	curPacket=p; // set temp var that most methods access.
	if(isRequest()) { // ICMP echo request.
		// 0 as last sqnr byte.
	    checkIfrestart();
	    handleEchoRequest();
	} else {
	    handleEchoReply();
	}
     }

    public void handleService(Service s ){
	System.out.println("Handle service: "+s);
	if(s.type()==ServiceConsts.WAKE_UP()) {// the time out occured
	    // maybe check also service nr.
	    gatherNsendResults();
	    if(series>0) {
		restart=true;
	    } else
		requestStopService(); // If all series are measured, then stop.
	}
    }

    public void emergencyHandle(IPPacket p) {}

    public boolean init(Vector initVector) {
	boolean ret=super.init(initVector);
	changeChannel("speedy.unibe.ch", ProtoConsts.CALLBACK_PORT,false,false);
	return ret;
    }

    public void stop() {}
    

    protected void handleEchoRequest() {
	if(identifierOk()) 
	    requestCntr.newArrival( curPacket.timeStamp,  sequenceNr());
    }

    protected void handleEchoReply() {
	if(identifierOk()) {
	    int seqnr=sequenceNr();
	    if(!requestCntr.beyond())
		if(seqnr>=requestCntr.expectedSeqNr()) // If reply is earlier than the request...
		    replyAnomaly++;
	    replyCntr.newArrival(curPacket.timeStamp, seqnr );
	}
    }

    protected void initForNewSerie() {
	replyAnomaly=0;
	int totTimeOut=serieLen*pingDistance +pingTimeOut;
	wakeupCall(totTimeOut, 1); // To evaluate the serie.
	series--;
    }
	

    protected void initSerieWithCurPacket() {
	identifier=identifier();
	int sqnr=sequenceNr();
	// Init the counters:
	requestCntr=new PingSeriesCounter(serieLen ,sqnr);
	replyCntr=new PingSeriesCounter(serieLen ,sqnr);
    }

    protected void checkIfrestart() {
	// Sync the agents on a special sequence nr, and avoid overflow problems
	if(!restart) return; // If we are still within a serie, do nothing.
	int seqnr=sequenceNr();
	if(isSyncNr(seqnr)) // Only start at special numbers, to be in sync with other agents.
	    if(isNotOverflowing(seqnr)) { // Start a new serie:
		restart=false; 
		initForNewSerie();
		initSerieWithCurPacket();
	    }
    }

    protected boolean isSyncNr(int seqnr) {
	return (seqnr%syncOffset==0);
    }

    protected boolean isNotOverflowing(int seqnr) {
	// Overflowing seq numbers are not handled nicely, thus we avoid them.
	return seqnr<(65535-serieLen);
    }

    protected void gatherNsendResults() {
	// Set the desired functionality here:
	//collectDelayNAlarm();
	collectJitterNAlarm(true);
    }

    protected void collectDelayNAlarm() {
	// We send a NumericResults with the series starting time and the average delay of the series.
	// It contains all delays measured in the series.
	// It is an alarm if eche replies were lost.
	Vector times=new Vector();
	NumericResults nv;
	long[] requeT=requestCntr.arrival;
	long[] replyT=replyCntr.arrival;
	long t1,t2, del;
	double sum=0;
	int divisor=serieLen;

	for(int i=0;i<serieLen;i++) {// Store the delays and calc the average
	    t1=requeT[i];
	    t2=replyT[i];
	    if((t1<0) || (t2<0)) { // Do not consider lost pings.
		divisor--;
	    } else {
		del=t2-t1;
		if(del<0) System.err.println("Warning: negative delay!");
		sum+=del;
		nv=new NumericResults(t1,del ); // Store the delay.
		times.addElement(nv);
	    }
	}

	nv=new NumericResults(requeT[0], times); 
	if(divisor>0)
	    nv.result=sum/divisor; // Store the average delay.
	else
	    nv.result=0;
	checkNsetAlarm(nv);
	sendResults(nv); // Send these results home.
    }


    protected void collectJitterNAlarm(boolean request) {
	// We send a NumericResults with the series starting time and the standard deviation (jitter) of the delay of the series.
	// if request==true we consider only the requests else the only the replies.
	// It contains all delay jitter  measured in the series (concrete delay minus average delay).
	// It also sets an alarm if duplicat,loss or reordering happens.
	Vector results=new Vector();
	NumericResults nv;
	long[] times;
	if(request)
	    times=requestCntr.arrival;
	else
	    times=replyCntr.arrival;
	long t, del,told=-1;
	double avrg, sum=0;
	int cntr=0;
      
	for(int i=0;i<serieLen;i++) {// Store the valid delays and calc the average
	    t=times[i];
	    if((t>=0) && (told>=0)) { // Do not consider lost pings.
		del=t-told;
		if(del<0) System.err.println("Warning: negative delay!");
		sum+=del;
		cntr++;
		nv=new NumericResults(t,del ); // Store the delay.
		results.addElement(nv);
	    }
	    told=t;
	}
	if(cntr>0)
	    avrg=sum/cntr; // Store the average delay.
	else
	    avrg=0;

	nv=new NumericResults(times[0], results); 
	nv.result=calculateStdDeviation(results, avrg);
	checkNsetAlarm(nv);
	sendResults(nv); // Send these results home.
    }

    protected double calculateStdDeviation(Vector delays, double avrg) {
	NumericResults nr;
	double del, dev, sum=0;
	
	for(Enumeration e=delays.elements(); e.hasMoreElements(); ) {
	    nr=(NumericResults) e.nextElement();
	    del=nr.result;
	    dev=del-avrg;
	    nr.result=dev;
	    sum+=dev*dev;
	}
	int degOfFree=delays.size()-1;
	if(degOfFree>0) {
	    sum=sum/degOfFree;
	    return Math.sqrt(sum);
	}
	return 0;
    }

    protected void checkNsetAlarm(NumericResults nr) {
	StringBuffer buf=new StringBuffer();
	if(replyAnomaly>0) {
	    buf.append("Replies earlier than requests: ");
	    buf.append(replyAnomaly);
	    buf.append('\n');
	}
	int problem=replyCntr.loss();
	if(problem>0) {
	    buf.append("Packet loss detected:\nLost ");
	    buf.append(problem);
	    buf.append(" out of: ");
	    buf.append(serieLen);
	}
	problem=replyCntr.duplicates();
	if(problem>0) {
	    buf.append("Packet duplicate detected:\nDuplicates: ");
	    buf.append(problem);
	}
	if(buf.length()>0)
	    nr.setAlarm(new String(buf));
    }

    protected boolean isRequest() {
	return (type()==8);
    }

    protected boolean identifierOk() {
	return (identifier()==identifier);
    }
    protected byte type() {
	return curPacket.dataByteAt(0);
    }

    protected int identifier() {
	return curPacket.toNumber(24,2);
    }
    protected int sequenceNr() {
	return curPacket.toBigEndianNumber(26,2); // if ping runs on a intel pc.
	//return curPacket.toNumber(26,2); // if ping runs on suns.
    }
 
}
