package clientserver;

import java.io.*;
import java.net.*;

import utils.*;

public  class ProtocolObjectSenderReceiver {
    // Provides servicing methods to send and receive(decode) protocol objects.
    // Authentication and type casting come to mind...
    // Decoding means e.g. decryption but could also be inflation. It
    // concers the PO field message, which is converted into a ByteArray
    // which represents a Message object.
 
    protected ErrorMessage error=null; // Stores error message to be sent.
    protected Socket socket;
    protected PGPEncoder263i pgp=new PGPEncoder263i("./","NodeCustomer"); // Note: writes tmp files into the current dir.
    protected String receiverID; // to encrypt, the receiver id must be known.
    protected boolean finished=false; // Can be used by subclasses to indicate, that
    // the connection was closed.
    protected ProtocolObject poFromWire; // Tmp store of received po, used for forwarding.

    public ProtocolObjectSenderReceiver(Socket s, String recId) {
	socket=s;
	receiverID=recId;
    }

    public synchronized boolean transit(ProtocolObject object) {
	// Transits (forwarding of agents) does not need new encoding.
	System.out.println("Transiting: "+object);
	return pureSend(object);
    }

    public synchronized boolean send(ProtocolObject object, String passPhrase) {
	// Sends the protocol object over the socket. Return true if things went well.

	if(!encode(object, passPhrase)) {
	    System.err.println("Could't encode protocol object!");
	    return false;
	}

	return pureSend(object);
    }

    protected boolean pureSend(ProtocolObject object) {
	try {
	    ObjectOutputStream out;
	    out=new ObjectOutputStream(socket.getOutputStream());
	    out.writeObject(object);
	}  catch (Exception e) {
	    System.err.println("Sending error while sending object: "+object.toString() + " "+e);
	    return false;
	}
	return true;
    }

    public synchronized DecodingResult receive(String passPhrase) {
	// Receive, create PO, sanity check, authentication, decode
	error=null;
        poFromWire=readFromSocket();
	if(error != null) {
	    return null;
	}
	DecodingResult res=decode(poFromWire, passPhrase);
	if(error != null) {
	    return null;
	}
	return res;
    }

    protected synchronized ProtocolObject readFromSocket() {
	try {
	    ObjectInputStream in;
	    in=new ObjectInputStream(socket.getInputStream());
	    return (ProtocolObject) in.readObject();
	} catch (Exception e) {
	    error=new ErrorMessage("Could not receive protocol object.");
	    System.err.println("Didn't receive the protocol object! "+e);
	    return null;
	}
    } 

    public boolean encode(ProtocolObject o, String pass) {
	// Returns true if a potential encryption of the protocol object was successfull.
	switch (o.encoding) {
	case ProtoConsts.PLAIN: {
	    return true; // no special encoding.
	}
	case ProtoConsts.FULL_PGP_v263i: {
	    InputStream enc=
		pgp.encode(o.message.getInputStream(), o.senderID, receiverID, pass);
	    if (enc==null) return false;
	    o.message.readByteCodeFromStream(enc); // Re-read the byte array from the encoded stream.
	    return true;
	}
	case ProtoConsts.SIGN_PGP_v263i: {
	    InputStream enc=
		pgp.sign(o.message.getInputStream(), o.senderID, pass);
	    if (enc==null) return false;
	    o.message.readByteCodeFromStream(enc); // Re-read the byte array from the encoded stream.
	    return true;
	}
	default: {
	    System.err.println("Warning: Unknown encoding demanded! Using plain text instead!");
	}
	return true;
	}
    }

    public DecodingResult decode(ProtocolObject wire, String pass) {
	byte thisEnc=wire.encoding;
	if(thisEnc==ProtoConsts.PLAIN)
	    return new DecodingResult(wire); // no special encoding.
	ProtocolObject o=new ProtocolObject(wire);
	DecodingResult ret=new DecodingResult(o);
	if((thisEnc==ProtoConsts.FULL_PGP_v263i) || (thisEnc==ProtoConsts.SIGN_PGP_v263i)) {
	    OneTimeInputStream dec=pgp.decode(o.message.getInputStream(), pass); 
	    ret=new DecodingResult(o,pgp.authenticate(dec), pgp.wasEncrypted(), pgp.getSigningParty(dec));
	    if (dec==null) error=new ErrorMessage("Decoding failed for: "+o); 
	    else
		o.message.readByteCodeFromStream(dec); // Re-read the byte array from the decoded stream.
	    return ret;
	} 
	// Otherwise:
	System.err.println("Warning: Received a message in unknown encoding!");
	error=new ErrorMessage("I don't know the message encoding type: "+o.encoding);
	return ret; 
    }

    public void closeSocket() {
	// The agent can close the socket when its finished via the ReplyChannel.
	try {
	    socket.close();
	    finished=true;
	} catch (IOException e) {
	    System.err.println(e);
	}
    }

}
