package stat;

public class X2  {
    // Dynamically recalculating X2 statistic.

    protected TabulatedCount count;
    protected DiscreteDistribution probs; // The probs of each category
    protected int nrOfCategories;
    protected int degreesOfFreedom;
    protected double V=0.0;
    protected double S=0.0;
    protected double I=0.0;

    protected int minNrOfMeasurements; // Only after this many measurements the
    // chi test is sufficiently approximated.
    
    public X2( DiscreteDistribution p, int d) {
	degreesOfFreedom=d;
	probs=p;
	count=new TabulatedCount(d+1);
	initMinNrOfMeasurements();

	int nrP=probs.getNrOfCategories(); // Categories equal?
	int nrC=count.getNrOfCategories();
	if (nrP != nrC) {
	    int min =nrP;
	    if (nrC<min) min=nrC; 
	    System.out.println("Warning: too many/little categories! Using the minimum: "+min);
	    nrOfCategories=min;
	} else {nrOfCategories=nrP;}
    }


    public void onlyAddMeasurement(int category) {
	// For later recalculation
	count.incCat(category);
    }


    public void newMeasurement(int category) {
	// For streamline measurements
	count.incCat(category);
	recalcV(category);
    }

    public int classify(byte m) {
	// changed for conformity with CPP sender
	//return ((int) m)+128;
	if (m<0) return 256+m;
	return m;
    }

    public int classify(int m) {
	return m;
    }


    protected void recalcV(int i) {
	long n=count.getTotal();
	I+=(2*count.getCountOf(i)-1)/probs.getProbabilityFor(i);
	S=I/(double) n;
	V=S-n;
    }

    public double getV() {return V;}

    public double getMinNrOfMeasurements() {
	return minNrOfMeasurements;
    }

    protected void initMinNrOfMeasurements() {
	// The expected value in each categrie should be bigger than 5
	double n=5/probs.min(); //Knuth's rule of thumb
	if (n>Integer.MAX_VALUE) {System.err.println("Min Prob. in distribution too small: \n -> too many measurements required for chi-square test.");
	}
	minNrOfMeasurements=(int) n;
    }



     public double calculateV() {
	 // For offline calculation
 	V=0;
	long c; // was int: a bug! the square of this will overflow!
	long n=count.getTotal();
	for (int s=0; s< nrOfCategories; s++) {
	    c=count.getCountOf(s);
	    V+=c*c/probs.getProbabilityFor(s);
	}
	V=V/n;
	V-=n;
	return V;
    }

    // --- Methods to evaluate V ---
    public boolean approximationOk() {
	return (minNrOfMeasurements<count.getTotal());
    }

    public boolean succeeds(boolean verySignificant) {
	// Check if the data succeeds the test.
	// Significance levels are 2% (verySignificant=true) or 10%.
	int up,low;
	if (verySignificant) {
	    up=6;low=0;
	} else {
	    up=5;low=1;
	}
	return ((V<percentagePoints(degreesOfFreedom,up))
		&& (V>percentagePoints(degreesOfFreedom,low)));
    }

    public boolean isExtremeV(double factor) {
	// Extimation of extremness of the V value. 
	// Use factors larger than 1.
	return ((V<percentagePoints(degreesOfFreedom,0)/factor)
|| (V>percentagePoints(degreesOfFreedom,6)*factor));
    }

    public double[] within() {
	double low=0.0;
	double up=1.0;
	for (int i=0; i<7; i++) {
	    up=X2Table.probVal[i];
	    if(V<percentagePoints(degreesOfFreedom,i)) break;
	    low=up;
	}
	if(low==up) up=1.0;
	double[] res=new double[2];
	res[0]=low;
	res[1]=up;
	return res;
    }

    // -- Output

    public String toString() {
	StringBuffer b=new StringBuffer();
	b.append("Chi-square test\n----------\nClassification: ");
	b.append(classify()); // The results
	b.append("\nCategory count:\n");
	b.append(count);
	b.append("\nDistribution:\n");
	b.append(probs);
	return new String(b);
    }

    public String classify() {
	StringBuffer b=new StringBuffer();
	b.append("V: "+V);
	double[] w= within();
	b.append(" lying in between "+w[0]);
	b.append(" and "+w[1]);
	return new String(b);
    }


    // --- Classification;

    protected static double percentagePoints(int degreesOfFreedom, int p) {
	//	System.err.println("dF: "+degreesOfFreedom);
	//Object stop=new Integer(1); // to raise an error
	if ((p<0) || (p>6)) {
	    System.err.println("Chi-square test only supported for significance level 0-7!");
	    return -1.0;
	}
	if (degreesOfFreedom>0) {
	    if(degreesOfFreedom<11) {
		return X2Table.percentPoints[degreesOfFreedom-1][p]; // Lookup in const table. (array counts from 0);
	    }
	    if(degreesOfFreedom>30)
		return calcApprox(degreesOfFreedom,p);
	}
	System.err.println("This Chi-square test only supportes freedom degrees of {1..10} and >30, not "+degreesOfFreedom);
	//System.out.println((Thread) stop); //raise an error
	return -1;
    }


    protected static double calcApprox(int degrOfFree, int p) {
	// Calculates the percentPoints for a degree of freedom higher than 30.
	double xp=X2Table.xp[p];
	return degrOfFree + Math.sqrt(2*degrOfFree)*xp + xp*xp*2.0/3.0 - 2.0/3.0;
	// The approx error is O(1/sqrt(degrOfFree) which is comparatively small. 
	// Tested and compared with the table in the literature for degOFree 30 and 50
    }

    // ----

//     public static void main(String[] args) {
// 	//DiscreteUniformDistribution d= new DiscreteUniformDistribution(12);
// 	//TabulatedCount measurement=new TabulatedCount(new FormatedSingletonFile("dicePerf.txt"),144);
// 	DiscreteDistribution dist= new DiscreteUniformDistribution(256);

// 	X2 x2=new X2(dist, 255);

// 	double rnd=0;
// 	int trunc=0;
// 	byte by=0;
// 	for (int i=0; i<10000; i++) {
// 	    rnd=Math.random();
// 	    rnd*=256;
// 	    trunc=(int) rnd;
// 	    trunc-=128;
// 	    by=(byte) trunc;
// 	    x2.newMeasurement(x2.classify(by));
// 	}

// 	System.out.println("Streamlined: "+x2.getV()); // Test against
// 	x2.calculateV();           //
//  	System.out.println(x2); // Offline version. test ok!

// // 	System.out.println(x2.succeeds(true));
// // 	System.out.println(x2.succeeds(false));
//     }
}
